import os
import pytest

from gitlabform.gitlab import AccessLevel
from gitlabform.gitlab.core import NotFoundException, UnexpectedResponseException
from tests.acceptance import run_gitlabform, DEFAULT_README


@pytest.fixture(scope="function")
def branches(request, gitlab, group_and_project):
    branches = [
        "protected_branch1",
        "protected_branch2",
        "protected_branch3",
        "regular_branch1",
        "regular_branch2",
    ]
    for branch in branches:
        gitlab.create_branch(group_and_project, branch, "main")
        if branch.startswith("protected"):
            gitlab.unprotect_branch(group_and_project, branch)
            gitlab.branch_access_level(
                group_and_project,
                branch,
                {
                    "push_access_level": AccessLevel.MAINTAINER.value,
                    "merge_access_level": AccessLevel.MAINTAINER.value,
                    "unprotect_access_level": AccessLevel.MAINTAINER.value,
                },
            )

    def fin():
        for branch in branches:
            gitlab.delete_branch(group_and_project, branch)

        params = (
            group_and_project,
            "main",
            "README.md",
            DEFAULT_README,
            "Reset default content",
        )
        try:
            gitlab.set_file(*params)
        except UnexpectedResponseException:
            gitlab.add_file(*params)

        gitlab.unprotect_branch(group_and_project, "main")
        gitlab.branch_access_level(
            group_and_project,
            "main",
            {
                "push_access_level": AccessLevel.MAINTAINER.value,
                "merge_access_level": AccessLevel.MAINTAINER.value,
                "unprotect_access_level": AccessLevel.MAINTAINER.value,
            },
        )

    request.addfinalizer(fin)


@pytest.fixture(scope="function")
def no_access_branch(request, gitlab, group_and_project):
    gitlab.create_branch(group_and_project, "no_access_branch", "main")
    gitlab.branch_access_level(
        group_and_project,
        "no_access_branch",
        {
            "push_access_level": AccessLevel.NO_ACCESS.value,
            "merge_access_level": AccessLevel.NO_ACCESS.value,
            "unprotect_access_level": AccessLevel.MAINTAINER.value,
        },
    )

    def fin():
        gitlab.unprotect_branch(group_and_project, "no_access_branch")
        gitlab.delete_branch(group_and_project, "no_access_branch")

    request.addfinalizer(fin)


@pytest.fixture(scope="class")
def file(request):
    f = open("file.txt", "a")
    f.write("Hanzi (Simplified): 丏丅丙两\nHanzi (Traditional): 丕不丈丁")
    f.close()

    def fin():
        os.remove("file.txt")

    request.addfinalizer(fin)


@pytest.fixture(scope="class")
def file2(request):
    f = open("file2.txt", "a")
    f.write("LICENSE\n\n“This is a license file”\n")
    f.close()

    def fin():
        os.remove("file2.txt")

    request.addfinalizer(fin)


class TestFiles:
    def test__set_file_specific_branch(self, gitlab, group_and_project, branches):

        set_file_specific_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              main:
                protected: true
                developers_can_push: false
                developers_can_merge: true
            files:
              "README.md":
                overwrite: true
                branches:
                  - main
                content: "Content for main only"
        """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, "main")
        assert commit["message"] == "Automated change made by gitlabform"

        file_content = gitlab.get_file(group_and_project, "main", "README.md")
        assert file_content == "Content for main only"

        other_branch_file_content = gitlab.get_file(
            group_and_project, "protected_branch1", "README.md"
        )
        assert other_branch_file_content == DEFAULT_README

        # check if main stays protected after the file update
        branch = gitlab.get_branch(group_and_project, "main")
        assert branch["protected"] is True

    def test__set_file_strongly_protected_branch(
        self, gitlab, group_and_project, no_access_branch
    ):

        set_file_specific_branch = f"""
            projects_and_groups:
              {group_and_project}:
                branches:
                  no_access_branch:
                    protected: true
                    push_access_level: 0 # no one
                    merge_access_level: 0 # no one
                    unprotect_access_level: 40
                files:
                  "README.md":
                    overwrite: true
                    branches:
                      - no_access_branch
                    content: "Content for no_access_branch only"
            """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, "no_access_branch")
        assert commit["message"] == "Automated change made by gitlabform"

        file_content = gitlab.get_file(
            group_and_project, "no_access_branch", "README.md"
        )
        assert file_content == "Content for no_access_branch only"

        other_branch_file_content = gitlab.get_file(
            group_and_project, "main", "README.md"
        )
        assert other_branch_file_content == DEFAULT_README

        # check if no_access_branch stays protected after the file update
        branch = gitlab.get_branch(group_and_project, "no_access_branch")
        assert branch["protected"] is True

    def test__delete_file_specific_branch(self, gitlab, group_and_project, branches):

        set_file_specific_branch = f"""
            projects_and_groups:
              {group_and_project}:
                branches:
                  main:
                    protected: true
                    developers_can_push: false
                    developers_can_merge: true
                files:
                  "README.md":
                    branches:
                      - main
                    delete: true
            """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, "main")
        assert commit["message"] == "Automated delete made by gitlabform"

        with pytest.raises(NotFoundException):
            gitlab.get_file(group_and_project, "main", "README.md")

        # check if main stays protected after the file delete
        branch = gitlab.get_branch(group_and_project, "main")
        assert branch["protected"] is True

    def test__custom_commit_message(self, gitlab, group_and_project, branches):

        set_file_specific_branch = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              main:
                protected: false
            files:
              "README.md":
                overwrite: true
                branches:
                  - main
                skip_ci: true
                content: "Hello world!"
                commit_message: "Preconfigured commit message"
        """

        run_gitlabform(set_file_specific_branch, group_and_project)

        commit = gitlab.get_last_commit(group_and_project, "main")
        assert commit["message"] == "Preconfigured commit message [skip ci]"

    def test__set_file_all_branches(self, gitlab, group_and_project, branches):

        set_file_all_branches = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              main:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protected_branch1:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protected_branch2:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protected_branch3:
                protected: true
                developers_can_push: false
                developers_can_merge: true
            files:
              "README.md":
                overwrite: true
                branches: all
                content: "Content for all branches"
        """
        run_gitlabform(set_file_all_branches, group_and_project)

        for branch in [
            "main",
            "protected_branch1",
            "protected_branch2",
            "protected_branch3",
            "regular_branch1",
            "regular_branch2",
        ]:
            file_content = gitlab.get_file(group_and_project, branch, "README.md")
            assert file_content == "Content for all branches"

        # check if these remain unprotected
        # (main branch is protected by default)
        for branch in [
            "regular_branch1",
            "regular_branch2",
        ]:
            branch = gitlab.get_branch(group_and_project, branch)
            assert branch["protected"] is False

    def test__set_file_protected_branches(self, gitlab, group_and_project, branches):

        set_file_protected_branches = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              main:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protected_branch1:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protected_branch2:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protected_branch3:
                protected: true
                developers_can_push: false
                developers_can_merge: true
            branches:
              protected_branch1:
                protected: true
                developers_can_push: true
                developers_can_merge: true
              protected_branch2:
                protected: true
                developers_can_push: true
                developers_can_merge: true
              protected_branch3:
                protected: true
                developers_can_push: true
                developers_can_merge: true
            files:
              "README.md":
                overwrite: true
                branches: protected
                content: "Content for protected branches only"
        """

        run_gitlabform(set_file_protected_branches, group_and_project)

        for branch in [
            "main",  # main branch is protected by default
            "protected_branch1",
            "protected_branch2",
            "protected_branch3",
        ]:
            file_content = gitlab.get_file(group_and_project, branch, "README.md")
            assert file_content == "Content for protected branches only"
            branch = gitlab.get_branch(group_and_project, branch)
            assert branch["protected"] is True

        for branch in ["regular_branch1", "regular_branch2"]:
            file_content = gitlab.get_file(group_and_project, branch, "README.md")
            assert file_content == DEFAULT_README
            branch = gitlab.get_branch(group_and_project, branch)
            assert branch["protected"] is False

    def test_set_file_protected_branches_new_api(self, gitlab, group_and_project):

        test_config = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              main:
                protected: true
                push_access_level: {AccessLevel.MAINTAINER.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        
            files:
              anyfile1:
                overwrite: true
                branches:
                  - main
                skip_ci: true
                content: foobar
        """

        run_gitlabform(test_config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "main", "anyfile1")
        assert file_content == "foobar"

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "main")
        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

    def test_set_file_protected_branches_new_api_not_all_levels(
        self, gitlab, group_and_project, branches
    ):

        test_config = f"""
            projects_and_groups:
              {group_and_project}:
                branches:
                  regular_branch1:
                    protected: true
                    # use a level that will require unprotecting the branch
                    # to change the file value
                    push_access_level: {AccessLevel.MAINTAINER.value}
                    merge_access_level: {AccessLevel.MAINTAINER.value}

                files:
                  anyfile2:
                    overwrite: true
                    branches:
                      - regular_branch1
                    content: barfoo
            """

        run_gitlabform(test_config, group_and_project)

        file_content = gitlab.get_file(group_and_project, "regular_branch1", "anyfile2")
        assert file_content == "barfoo"

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "regular_branch1")
        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        # the default value
        # according to https://docs.gitlab.com/ee/api/protected_branches.html#protect-repository-branches
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

    def test__set_file_with_chinese_characters(self, gitlab, group, project, branches):
        group_and_project_name = f"{group}/{project}"

        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project_name}:
            files:
              "README.md":
                overwrite: true
                branches:
                  - main
                content: |
                    Hanzi (Traditional): 丕不丈丁
                    Hanzi (Simplified): 丏丅丙两
        """

        run_gitlabform(set_file_chinese_characters, group_and_project_name)

        file_content = gitlab.get_file(group_and_project_name, "main", "README.md")
        assert file_content == "Hanzi (Traditional): 丕不丈丁\nHanzi (Simplified): 丏丅丙两"

    def test__set_external_file_with_chinese_characters(
        self, gitlab, group, project, branches, file
    ):
        group_and_project_name = f"{group}/{project}"

        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project_name}:
            files:
              "README.md":
                overwrite: true
                branches:
                  - main
                file: file.txt
        """

        run_gitlabform(set_file_chinese_characters, group_and_project_name)

        file_content = gitlab.get_file(group_and_project_name, "main", "README.md")
        assert file_content == "Hanzi (Simplified): 丏丅丙两\nHanzi (Traditional): 丕不丈丁"

    def test__set_external_file_with_utf8_characters(
        self, gitlab, group, project, branches, file2
    ):
        group_and_project_name = f"{group}/{project}"

        set_file_chinese_characters = f"""
        projects_and_groups:
          {group_and_project_name}:
            files:
              "README.md":
                overwrite: true
                branches:
                  - main
                file: file2.txt
        """

        run_gitlabform(set_file_chinese_characters, group_and_project_name)

        file_content = gitlab.get_file(group_and_project_name, "main", "README.md")
        assert file_content == "LICENSE\n\n“This is a license file”"



from logging import debug
from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class GroupSettingsProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("group_settings", gitlab)

    def _process_configuration(self, group: str, configuration: dict):
        group_settings = configuration["group_settings"]
        debug("Group settings BEFORE: %s", self.gitlab.get_group_settings(group))
        verbose(f"Setting group settings: {group_settings}")
        self.gitlab.put_group_settings(group, group_settings)
        debug("Group settings AFTER: %s", self.gitlab.get_group_settings(group))



from abc import ABC, abstractmethod


class AbstractKey(ABC):
    @abstractmethod
    def matches(self, e1, e2):
        pass

    @abstractmethod
    def contains(self, entity):
        pass

    @abstractmethod
    def explain(self) -> str:
        pass


class Key(AbstractKey):
    def __init__(self, value):
        self.value = value

    def matches(self, e1, e2):
        return (
            self.value in e1 and self.value in e2 and e1[self.value] == e2[self.value]
        )

    def contains(self, entity):
        return entity.get(self.value, None) is not None

    def explain(self) -> str:
        return f"'{self.value}'"


class And(AbstractKey):
    def __init__(self, *arg: AbstractKey):
        self.keys = arg

    def matches(self, e1, e2):
        return all([key.matches(e1, e2) for key in self.keys])

    def contains(self, entity):
        return all([key.contains(entity) for key in self.keys])

    def explain(self) -> str:
        explains = [key.explain() for key in self.keys]
        return f"({' and '.join(explains)})"


class Or(AbstractKey):
    def __init__(self, *arg: AbstractKey):
        self.keys = arg

    def matches(self, e1, e2):
        return any([key.matches(e1, e2) for key in self.keys])

    def contains(self, entity):
        return any([key.contains(entity) for key in self.keys])

    def explain(self) -> str:
        explains = [key.explain() for key in self.keys]
        return f"({' or '.join(explains)})"


class Xor(AbstractKey):
    def __init__(self, *arg: AbstractKey):
        self.keys = arg

    # copied from https://stackoverflow.com/a/16801336/2693875
    @staticmethod
    def _single_true(iterable):
        iterator = iter(iterable)

        # consume from "i" until first true or it's exhausted
        has_true = any(iterator)

        # carry on consuming until another true value / exhausted
        has_another_true = any(iterator)

        # True if exactly one true found
        return has_true and not has_another_true

    def matches(self, e1, e2):
        return self._single_true([key.matches(e1, e2) for key in self.keys])

    def contains(self, entity):
        return self._single_true([key.contains(entity) for key in self.keys])

    def explain(self) -> str:
        explains = [key.explain() for key in self.keys]
        return f"(exactly one of: {', '.join(explains)})"



from logging import debug
from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class DeployKeysProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("deploy_keys", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        debug("Deploy keys BEFORE: %s", self.gitlab.get_deploy_keys(project_and_group))
        for deploy_key in sorted(configuration["deploy_keys"]):
            verbose(f"Setting deploy key: {deploy_key}")
            self.gitlab.post_deploy_key(
                project_and_group, configuration["deploy_keys"][deploy_key]
            )
        debug("Deploy keys AFTER: %s", self.gitlab.get_deploy_keys(project_and_group))



import json
from logging import debug
from cli_ui import debug as verbose

import pkg_resources
import requests
import os

# noinspection PyPackageRequirements
import urllib3

# noinspection PyPackageRequirements
from urllib3.util.retry import Retry
from urllib import parse
from requests.adapters import HTTPAdapter

from gitlabform.configuration import Configuration


class GitLabCore:
    def __init__(self, config_path=None, config_string=None):

        self.configuration = Configuration(config_path, config_string)

        self.url = self.configuration.get("gitlab|url", os.getenv("GITLAB_URL"))
        self.token = self.configuration.get("gitlab|token", os.getenv("GITLAB_TOKEN"))
        self.ssl_verify = self.configuration.get("gitlab|ssl_verify", True)
        self.timeout = self.configuration.get("gitlab|timeout", 10)

        self.session = requests.Session()

        retries = Retry(
            total=3, backoff_factor=0.25, status_forcelist=[500, 502, 503, 504]
        )

        self.session.mount("http://", HTTPAdapter(max_retries=retries))
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

        self.session.verify = self.ssl_verify
        if not self.ssl_verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self.gitlabform_version = pkg_resources.get_distribution("gitlabform").version
        self.requests_version = pkg_resources.get_distribution("requests").version
        self.session.headers.update(
            {
                "private-token": self.token,
                "authorization": f"Bearer {self.token}",
                "user-agent": f"GitLabForm/{self.gitlabform_version} (python-requests/{self.requests_version})",
            }
        )

        try:
            version = self._make_requests_to_api("version")
            verbose(
                f"Connected to GitLab version: {version['version']} ({version['revision']})"
            )
            self.version = version["version"]
        except Exception as e:
            raise TestRequestFailedException(e)

    def get_configuration(self):
        return self.configuration

    def get_project(self, project_and_group_or_id):
        return self._make_requests_to_api("projects/%s", project_and_group_or_id)

    def _get_user_id(self, username):
        users = self._make_requests_to_api("users?username=%s", username, "GET")

        # this API endpoint is for lookup, not search, so 'username' has to be full and exact username
        # also it's not possible to get more than 1 user as a result

        if len(users) == 0:
            raise NotFoundException(
                "No users found when searching for username '%s'" % username
            )

        return users[0]["id"]

    def _get_user(self, user_id):
        return self._make_requests_to_api("users/%s", str(user_id), "GET")

    def _get_group_id(self, path):
        group = self._make_requests_to_api("groups/%s", path, "GET")
        # TODO: add tests for all that uses this and then stop converting these ints to strings here
        return str(group["id"])

    def _get_project_id(self, project_and_group):
        # This is a NEW workaround for https://github.com/gitlabhq/gitlabhq/issues/8290
        result = self.get_project(project_and_group)
        return str(result["id"])

    def has_no_license(self):
        license = self._make_requests_to_api("license")
        return not license or license["expired"]

    def _make_requests_to_api(
        self,
        path_as_format_string,
        args=None,
        method="GET",
        data=None,
        expected_codes=200,
        paginated=False,
        json=None,
    ):
        """
        Makes a HTTP request (or requests) to the GitLab API endpoint. It takes case of making as many requests as
        needed in case we are using a paginated endpoint. (See underlying method for authentication, retries,
        timeout etc.)

        :param path_as_format_string: path with parts to be replaced by values from `args` replaced by '%s'
                                      (aka the old-style Python string formatting, see:
                                       https://docs.python.org/2/library/stdtypes.html#string-formatting )
        :param args: single element or a tuple of values to put under '%s's in `path_as_format_string`
        :param method: uppercase string of a HTTP method name, like 'GET' or 'PUT'
        :param data: dict with data to be 'PUT'ted or 'POST'ed
        :param expected_codes: a single HTTP code (like: 200) or a list of accepted HTTP codes
                               - if the call to the API will return other code an exception will be thrown
        :param paginated: if given API is paginated (see https://docs.gitlab.com/ee/api/#pagination )
        :param json: alternatively to `dict` you can set this to a string that can be parsed as JSON that will
                     be used as data to be 'PUT'ted or 'POST'ed
        :return: data returned by the endpoint, as a JSON object. If the API is paginated the it returns JSONs with
                 arrays of objects and then this method returns JSON with a single array that contains all of those
                 objects.
        """
        if not paginated:
            response = self._make_request_to_api(
                path_as_format_string, args, method, data, expected_codes, json
            )
            return response.json()
        else:
            if "?" in path_as_format_string:
                path_as_format_string += "&per_page=100"
            else:
                path_as_format_string += "?per_page=100"

            first_response = self._make_request_to_api(
                path_as_format_string, args, method, data, expected_codes, json
            )
            results = first_response.json()

            # In newer versions of GitLab the 'X-Total-Pages' may not be available
            # anymore, see https://gitlab.com/gitlab-org/gitlab/-/merge_requests/43159
            # so let's switch to thew newer style.

            response = first_response
            while True:
                if (
                    "x-next-page" in response.headers
                    and response.headers["x-next-page"]
                ):
                    next_page = response.headers["x-next-page"]
                    response = self._make_request_to_api(
                        path_as_format_string + "&page=" + str(next_page),
                        args,
                        method,
                        data,
                        expected_codes,
                        json,
                    )
                    results += response.json()
                else:
                    break

        return results

    def _make_request_to_api(
        self, path_as_format_string, args, method, dict_data, expected_codes, json_data
    ):

        """
        Makes a single request to the GitLab API. Takes care of the authentication, basic error processing,
        retries, timeout etc.

        :param for the params description please see `_make_requests_to_api()`
        :return: data returned by the endpoint, as a JSON object.
        """

        expected_codes = self._listify(expected_codes)

        if dict_data and json_data:
            raise Exception(
                "You need to pass the data either as dict (dict_data) or JSON (json_data), not both!"
            )

        url = (
            self.url
            + "/api/v4/"
            + self._format_with_url_encoding(path_as_format_string, args)
        )
        debug(
            "url = %s , method = %s , data = %s, json = %s",
            url,
            method,
            json.dumps(dict_data, sort_keys=True),
            json.dumps(json_data, sort_keys=True),
        )
        if dict_data:
            response = self.session.request(
                method, url, data=dict_data, timeout=self.timeout
            )
        elif json_data:
            response = self.session.request(
                method, url, json=json_data, timeout=self.timeout
            )
        else:
            response = self.session.request(method, url, timeout=self.timeout)
        debug("response code=%s" % response.status_code)

        if response.status_code in expected_codes:
            # if we accept error responses then they will likely not contain a JSON body
            # so fake it to fix further calls to response.json()
            if response.status_code == 204 or (400 <= response.status_code <= 499):
                debug("faking response body to be {}")
                response.json = lambda: {}
        else:
            if response.status_code == 404:
                raise NotFoundException(
                    f"Resource with url='{url}' not found (HTTP 404)!"
                )
            else:
                data_output = f"data={dict_data}" if dict_data else f"json={json_data}"
                raise UnexpectedResponseException(
                    f"Request url='{url}', method={method}, {data_output} failed -"
                    f" expected code(s) {str(expected_codes)},"
                    f" got code {response.status_code} & body: '{response.text}'",
                    response.status_code,
                    response.text,
                )

        debug("response json=%s" % json.dumps(response.json(), sort_keys=True))
        return response

    @staticmethod
    def _format_with_url_encoding(format_string, single_arg_or_args_tuple):

        # we want to URL-encode all the args, but not the path itself which looks like "/foo/%s/bar"
        # because '/'s here are NOT to be URL-encoded

        if not single_arg_or_args_tuple:
            # there are no params, so the format_string is the URL
            return format_string
        else:
            if type(single_arg_or_args_tuple) == tuple:
                # URL-encode each arg in the tuple and return it as tuple too
                url_encoded_args = ()
                for arg in single_arg_or_args_tuple:
                    url_encoded_args += (parse.quote_plus(str(arg)),)
            else:
                # URL-encode single arg
                url_encoded_args = parse.quote_plus(str(single_arg_or_args_tuple))

            return format_string % url_encoded_args

    @staticmethod
    def _listify(expected_codes):
        if isinstance(expected_codes, int):
            return [expected_codes]
        else:
            return expected_codes


class TestRequestFailedException(Exception):
    def __init__(self, underlying: Exception):
        self.underlying = underlying


class NotFoundException(Exception):
    pass


class TimeoutWaitingForDeletion(Exception):
    pass


class InvalidParametersException(Exception):
    pass


class UnexpectedResponseException(Exception):
    def __init__(self, message: str, response_status_code: int, response_text: str):
        self.message: str = message
        self.response_status_code: int = response_status_code
        self.response_text: str = response_text

    def __str__(self):
        return self.message



from cli_ui import warning

from gitlabform.gitlab import GitLab
from gitlabform.output import EffectiveConfiguration
from gitlabform.processors.abstract_processor import AbstractProcessor


from gitlabform.processors.util.decorators import configuration_to_safe_dict


# this processor exists only to prevent the configs with just `group_shared_with` key on a group
# level to not be processed because of an empty effective config
class GroupSharedWithProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("group_shared_with", gitlab)

    @configuration_to_safe_dict
    def process(
        self,
        project_or_project_and_group: str,
        configuration: dict,
        dry_run: bool,
        effective_configuration: EffectiveConfiguration,
    ):
        warning(
            "Using `group_shared_with:` is deprecated and will be removed in future versions "
            "of GitLabForm. Please move its contents to `group_members.groups`."
        )

    def _process_configuration(self, group: str, configuration: dict):
        pass



from cli_ui import debug as verbose
from cli_ui import warning

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor


class ServicesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("services", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        for service in sorted(configuration["services"]):
            if configuration.get("services|" + service + "|delete"):
                verbose(f"Deleting service: {service}")
                self.gitlab.delete_service(project_and_group, service)
            else:

                if (
                    "recreate" in configuration["services"][service]
                    and configuration["services"][service]["recreate"]
                ):
                    # support from this configuration key has been added in v1.13.4
                    # we will remove it here to avoid passing it to the GitLab API
                    warning(
                        f"Ignoring deprecated 'recreate' field in the '{service}' service config. "
                        "Please remove it from the config file permanently as this workaround is not "
                        "needed anymore."
                    )
                    del configuration["services"][service]["recreate"]

                verbose(f"Setting service: {service}")
                self.gitlab.set_service(
                    project_and_group, service, configuration["services"][service]
                )



import logging
import pytest

from gitlabform.configuration import Configuration

logger = logging.getLogger(__name__)


@pytest.fixture
def configuration_with_subgroups_and_projects():
    config_yaml = """
    ---
    projects_and_groups:
      some_group/*:
        project_settings:
          foo: bar
        hooks:
          a:
            foo: bar

      some_group/subgroup_level_1/*:
        project_settings:
          foo: bar2
        hooks:
          a:
            foo: bar2

      some_group/subgroup_level_1/subgroup_level_2/*:
        project_settings:
          foo: bar3
        hooks:
          a:
            foo: bar3

      some_group/some_project:
        project_settings:
          bar: something_else
        hooks:
          b:
            bar: something_else

      some_group/subgroup_level_1/some_project:
          project_settings:
            bar: something_else2
          hooks:
            b:
              bar: something_else2

      some_group/subgroup_level_1/subgroup_level_2/some_project:
          project_settings:
            bar: something_else3
          hooks:
            b:
              bar: something_else3
    """
    return Configuration(config_string=config_yaml)


@pytest.fixture
def configuration_with_subgroups():
    config_yaml = """
    ---
    projects_and_groups:
      some_group/*:
        group_members:
          my-user:
            access_level: 10
        enforce_group_members: true

      some_group/subgroup/*:
        group_members:
          my-user2:
            access_level: 20
        enforce_group_members: true
    """
    return Configuration(config_string=config_yaml)


def test__get_effective_config_for_project__project_from_config__level1(
    configuration_with_subgroups_and_projects,
):
    effective_config = (
        configuration_with_subgroups_and_projects.get_effective_config_for_project(
            "some_group/subgroup_level_1/some_project"
        )
    )

    additive__project_settings = effective_config["project_settings"]

    # project and only subgroup level 1
    assert additive__project_settings == {"foo": "bar2", "bar": "something_else2"}


def test__get_effective_config_for_project__project_from_config__level2(
    configuration_with_subgroups_and_projects,
):
    effective_config = (
        configuration_with_subgroups_and_projects.get_effective_config_for_project(
            "some_group/subgroup_level_1/subgroup_level_2/some_project"
        )
    )

    additive__project_settings = effective_config["project_settings"]

    # project and only subgroup level 2
    assert additive__project_settings == {"foo": "bar3", "bar": "something_else3"}


def test__get_effective_config_for_subgroup(
    configuration_with_subgroups,
):
    effective_config = configuration_with_subgroups.get_effective_config_for_group(
        "some_group/subgroup"
    )

    assert effective_config == {
        "group_members": {
            "my-user2": {
                "access_level": 20,
            },
        },
        "enforce_group_members": True,
    }



from logging import debug

from gitlabform.configuration.core import ConfigurationCore


class ConfigurationGroups(ConfigurationCore):
    def __init__(self, config_path=None, config_string=None):
        super().__init__(config_path, config_string)

    def get_groups(self) -> list:
        """
        :return: sorted list of groups that are EXPLICITLY defined in the config
        """
        groups = []
        projects_and_groups = self.get("projects_and_groups")
        for element in projects_and_groups.keys():
            if element.endswith("/*"):
                # cut off that "/*"
                group_name = element[:-2]
                groups.append(group_name)
        return sorted(groups)

    def get_effective_config_for_group(self, group) -> dict:
        """
        :param group: "group_name"
        :return: merged configuration for this group, from common, group. Merging is additive.
        """

        common_config = self.get_common_config()
        debug("Common config: %s" % common_config)

        group_config = self.get_group_config(group)
        debug("Group config: %s" % group_config)

        if not group_config and not common_config:
            return {}

        return self.merge_configs(common_config, group_config)

    def get_effective_subgroup_config(self, subgroup):

        #
        # Goes through a subgroups hierarchy, from top to bottom
        #
        # "x/y/x" -> ["x", "x/y", "x/y/z"]
        #
        # ...and for each element after 1st generate effective config from previous effective one merged with current:
        #
        #              |     v       |
        #              \---> a       |
        #                    |       v
        #                    \------>b = effective config to return
        #
        # ..where a = merged_config("x", "x/y") and b = merged_config(a, "x/y/z")
        #

        effective_config = {}
        elements = subgroup.split("/")
        last_element = None
        for element in elements:
            if not last_element:
                effective_config = self.get_group_config(element)
                debug("First level config for '%s': %s" % (element, effective_config))
                last_element = element
            else:
                next_level_subgroup = last_element + "/" + element
                next_level_subgroup_config = self.get_group_config(next_level_subgroup)
                debug(
                    "Config for '%s': %s"
                    % (next_level_subgroup, next_level_subgroup_config)
                )
                effective_config = self.merge_configs(
                    effective_config, next_level_subgroup_config
                )
                debug(
                    "Merged previous level config for '%s' with config for '%s': %s"
                    % (last_element, next_level_subgroup, effective_config)
                )
                last_element = last_element + "/" + element

        return effective_config


class EmptyConfigException(Exception):
    pass



from logging import debug
from cli_ui import debug as verbose
from cli_ui import warning

from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import NotFoundException
from gitlabform.processors.abstract_processor import AbstractProcessor


class GroupSecretVariablesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("group_secret_variables", gitlab)

    def _process_configuration(self, group: str, configuration: dict):
        debug(
            "Group secret variables BEFORE: %s",
            self.gitlab.get_group_secret_variables(group),
        )

        for secret_variable in sorted(configuration["group_secret_variables"]):

            if "delete" in configuration["group_secret_variables"][secret_variable]:
                key = configuration["group_secret_variables"][secret_variable]["key"]
                if configuration["group_secret_variables"][secret_variable]["delete"]:
                    verbose(f"Deleting {secret_variable}: {key} in group {group}")
                    try:
                        self.gitlab.delete_group_secret_variable(group, key)
                    except:
                        warning(f"Could not delete variable {key} in group {group}")
                    continue

            verbose(f"Setting group secret variable: {secret_variable}")
            try:
                self.gitlab.put_group_secret_variable(
                    group, configuration["group_secret_variables"][secret_variable]
                )
            except NotFoundException:
                self.gitlab.post_group_secret_variable(
                    group, configuration["group_secret_variables"][secret_variable]
                )

        debug(
            "Groups secret variables AFTER: %s",
            self.gitlab.get_group_secret_variables(group),
        )



import pytest
from tests.acceptance import (
    run_gitlabform,
    get_gitlab,
)


gl = get_gitlab()


class TestPushRules:
    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__create_push_rules(self, gitlab, group_and_project):
        config = f"""
        projects_and_groups:
          {group_and_project}:
            project_push_rules:
              commit_message_regex: 'Fixes \\d +'
              branch_name_regex: ""
              deny_delete_tag: false
              member_check: false
              prevent_secrets: false
              author_email_regex: ""
              file_name_regex: ""
              max_file_size: 0 # in MB, 0 means unlimited
        """

        run_gitlabform(config, group_and_project)

        push_rules = gitlab.get_project_push_rules(group_and_project)
        assert push_rules["max_file_size"] == 0

    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__edit_push_rules(self, gitlab, group_and_project):

        self.test__create_push_rules(gitlab, group_and_project)

        config = f"""
        projects_and_groups:
          {group_and_project}:
            project_push_rules:
              commit_message_regex: 'Fixes \\d +'
              branch_name_regex: ""
              deny_delete_tag: false
              member_check: false
              prevent_secrets: false
              author_email_regex: ""
              file_name_regex: ""
              max_file_size: 2 # in MB, 0 means unlimited
        """

        run_gitlabform(config, group_and_project)

        push_rules = gitlab.get_project_push_rules(group_and_project)
        assert push_rules["max_file_size"] == 2



# import pytest
#
# from gitlabform import GitLabForm
# from gitlabform.test import (
#     create_group,
#     get_gitlab,
#     delete_variables_from_group,
#     GROUP_NAME,
# )
#
#
# @pytest.fixture(scope="function")
# def gitlab(request):
#     create_group(GROUP_NAME)
#
#     gl = get_gitlab()
#
#     def fin():
#         delete_variables_from_group(GROUP_NAME, ["FOO", "BAR"])
#
#     request.addfinalizer(fin)
#     return gl  # provide fixture value
#
#
# config_single_secret_variable = f"""
# group_settings:
#   {GROUP_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     group_secret_variables:
#       foo:
#         key: FOO
#         value: 123
# """
#
# config_delete_secret_variable = f"""
# group_settings:
#   {GROUP_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     group_secret_variables:
#       foo:
#         key: FOO
#         value: 123
#         delete: true
# """
#
# config_single_secret_variable2 = f"""
# group_settings:
#   {GROUP_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     group_secret_variables:
#       foo:
#         key: FOO
#         value: 123456
# """
#
# config_more_secret_variables = f"""
# group_settings:
#   {GROUP_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     group_secret_variables:
#       foo:
#         key: FOO
#         value: 123456
#       bar:
#         key: BAR
#         value: bleble
# """
#
# config_masked_secret_variables = f"""
# group_settings:
#   {GROUP_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     group_secret_variables:
#       foo:
#         key: FOO
#         # https://docs.gitlab.com/ee/ci/variables/#masked-variable-requirements
#         value: 12345678
#         masked: true
# """
#
# config_protected_secret_variables = f"""
# group_settings:
#   {GROUP_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     group_secret_variables:
#       foo:
#         key: FOO
#         value: 123
#         protected: true
# """
#
#
# class TestGroupSecretVariables:
#     def test__single_secret_variable(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_group_secret_variables(GROUP_NAME)
#         assert len(secret_variables) == 1
#
#     def test__reset_single_secret_variable(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_group_secret_variables(GROUP_NAME)
#         assert len(secret_variables) == 1
#         assert secret_variables[0]["value"] == "123"
#
#         gf = GitLabForm(
#             config_string=config_single_secret_variable2, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_group_secret_variables(GROUP_NAME)
#         assert len(secret_variables) == 1
#         assert secret_variables[0]["value"] == "123456"
#
#     def test__delete_secret_variable(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_group_secret_variables(GROUP_NAME)
#         assert len(secret_variables) == 1
#         assert secret_variables[0]["value"] == "123"
#
#         gf = GitLabForm(
#             config_string=config_delete_secret_variable, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_group_secret_variables(GROUP_NAME)
#         assert len(secret_variables) == 0
#
#     def test__more_secret_variables(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_more_secret_variables, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_group_secret_variables(GROUP_NAME)
#         secret_variables_keys = set([secret["key"] for secret in secret_variables])
#         assert len(secret_variables) == 2
#         assert secret_variables_keys == {"FOO", "BAR"}
#
#     def test__masked_secret_variables(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_masked_secret_variables, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variable = gitlab.get_group_secret_variable_object(GROUP_NAME, "FOO")
#         assert secret_variable["value"] == "12345678"
#         assert secret_variable["masked"]
#
#     def test__protected_secret_variables(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_protected_secret_variables, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variable = gitlab.get_group_secret_variable_object(GROUP_NAME, "FOO")
#         assert secret_variable["value"] == "123"
#         assert secret_variable["protected"]
#
#     def test__protected_change_secret_variables(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variable = gitlab.get_group_secret_variable_object(GROUP_NAME, "FOO")
#         assert secret_variable["value"] == "123"
#         assert secret_variable["protected"] is False
#
#         gf = GitLabForm(
#             config_string=config_protected_secret_variables, project_or_group=GROUP_NAME
#         )
#         gf.main()
#
#         secret_variable = gitlab.get_group_secret_variable_object(GROUP_NAME, "FOO")
#         assert secret_variable["value"] == "123"
#         assert secret_variable["protected"] is True



from logging import debug
from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.difference_logger import DifferenceLogger


class ProjectSettingsProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("project_settings", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        project_settings = configuration["project_settings"]
        debug(
            "Project settings BEFORE: %s",
            self.gitlab.get_project_settings(project_and_group),
        )
        verbose(f"Setting project settings: {project_settings}")
        self.gitlab.put_project_settings(project_and_group, project_settings)
        debug(
            "Project settings AFTER: %s",
            self.gitlab.get_project_settings(project_and_group),
        )

    def _print_diff(self, project_and_group: str, project_settings):
        current_project_settings = self.gitlab.get_project_settings(project_and_group)
        DifferenceLogger.log_diff(
            "Project %s changes" % project_and_group,
            current_project_settings,
            project_settings,
        )



# import pytest
#
# from gitlabform.gitlabform import GitLabForm
# from gitlabform.gitlabform.test import (
#     create_group,
#     create_project_in_group,
#     create_users_in_project,
#     add_users_to_group,
#     get_gitlab,
#     GROUP_NAME,
#     OWNER_ACCESS,
#     remove_users_from_group,
# )
#
# PROJECT_NAME = "merge_requests_project"
# GROUP_AND_PROJECT_NAME = GROUP_NAME + "/" + PROJECT_NAME
#
# USER_BASE_NAME = "merge_requests_user"  # user1, user2, ...
#
# GROUP_WITH_USER1_AND_USER2 = "gitlabform_tests_group_with_user1_and_user2"
#
# GROUP_WITH_USER3 = "gitlabform_tests_group_with_user3"
#
# GROUP_WITH_USER4 = "gitlabform_tests_group_with_user4"
#
# DEVELOPER_ACCESS = 30
#
#
# @pytest.fixture(scope="function")
# def gitlab(request):
#
#     create_group(GROUP_NAME)
#     create_project_in_group(GROUP_NAME, PROJECT_NAME)
#
#     create_users_in_project(USER_BASE_NAME, 4, GROUP_AND_PROJECT_NAME)
#
#     create_group(GROUP_WITH_USER1_AND_USER2)
#     add_users_to_group(
#         GROUP_WITH_USER1_AND_USER2,
#         ["merge_requests_user1", "merge_requests_user2"],
#         OWNER_ACCESS,
#     )
#     remove_users_from_group(GROUP_WITH_USER1_AND_USER2, ["root"])
#
#     create_group(GROUP_WITH_USER3)
#     add_users_to_group(GROUP_WITH_USER3, ["merge_requests_user3"], OWNER_ACCESS)
#     remove_users_from_group(GROUP_WITH_USER3, ["root"])
#
#     create_group(GROUP_WITH_USER4)
#     add_users_to_group(GROUP_WITH_USER4, ["merge_requests_user4"], OWNER_ACCESS)
#     remove_users_from_group(GROUP_WITH_USER4, ["root"])
#
#     gl = get_gitlab()
#
#     def fin():
#         # remove all the approval rules and disable approvals in the project
#         rules = gl.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         for rule in rules:
#             gl.delete_approvals_rule(GROUP_AND_PROJECT_NAME, rule["id"])
#         gl.put_project_settings(
#             GROUP_AND_PROJECT_NAME, {"merge_requests_access_level": "disabled"}
#         )
#
#     request.addfinalizer(fin)
#     return gl  # provide fixture value
#
#
# config__approvers_single_user = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user1
# """
#
# config__approvers_single_user_change = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user2
# """
#
# config__approvers_single_group = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approver_groups:
#         - gitlabform_tests_group_with_user1_and_user2
# """
#
# config__approvers_single_group_change = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approver_groups:
#         - gitlabform_tests_group_with_user4
# """
#
# config__approvers_single_user_and_single_group = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user1
#       approver_groups:
#         - gitlabform_tests_group_with_user4
# """
#
# config__approvers_single_user_and_single_group_change = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user2
#       approver_groups:
#         - gitlabform_tests_group_with_user1_and_user2
# """
#
# config__approvers_more_than_one_user = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user1
#         - merge_requests_user2
# """
#
# config__approvers_more_than_one_user_and_more_than_one_group = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user1
#         - merge_requests_user2
#       approver_groups:
#         - gitlabform_tests_group_with_user3
#         - gitlabform_tests_group_with_user4
# """
#
#
# config__approvers_removing_preexisting_approvals = """
# project_settings:
#   gitlabform_tests_group/merge_requests_project:
#     project_settings:
#       merge_requests_access_level: "enabled"
#     merge_requests:
#       remove_other_approval_rules: true
#       approvals:
#         approvals_before_merge: 1
#         reset_approvals_on_push: true
#         disable_overriding_approvers_per_merge_request: true
#       approvers:
#         - merge_requests_user1
#       approver_groups:
#         - gitlabform_tests_group_with_user3
# """
#
#
# class TestMergeRequestApprovers:
#     def test__if_it_works__single_user(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_single_user,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["users"]) == 1
#         assert rules[0]["users"][0]["username"] == "merge_requests_user1"
#         assert len(rules[0]["groups"]) == 0
#
#     def test__if_change_works__only_users(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_single_user,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["users"]) == 1
#         assert rules[0]["users"][0]["username"] == "merge_requests_user1"
#         assert len(rules[0]["groups"]) == 0
#
#         gf = GitLabForm(
#             config_string=config__approvers_single_user_change,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["users"]) == 1
#         assert rules[0]["users"][0]["username"] == "merge_requests_user2"
#         assert len(rules[0]["groups"]) == 0
#
#     def test__if_it_works__single_group(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_single_group,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["groups"]) == 1
#         assert (
#             rules[0]["groups"][0]["name"]
#             == "gitlabform_tests_group_with_user1_and_user2"
#         )
#         assert len(rules[0]["users"]) == 0
#
#     def test__if_change_works__single_group_change(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_single_group,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["groups"]) == 1
#         assert (
#             rules[0]["groups"][0]["name"]
#             == "gitlabform_tests_group_with_user1_and_user2"
#         )
#         assert len(rules[0]["users"]) == 0
#
#         gf = GitLabForm(
#             config_string=config__approvers_single_group_change,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["groups"]) == 1
#         assert rules[0]["groups"][0]["name"] == "gitlabform_tests_group_with_user4"
#         assert len(rules[0]["users"]) == 0
#
#     def test__if_it_works__single_user_and_single_group(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_single_user_and_single_group,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["users"]) == 1
#         assert rules[0]["users"][0]["username"] == "merge_requests_user1"
#         assert len(rules[0]["groups"]) == 1
#         assert rules[0]["groups"][0]["name"] == "gitlabform_tests_group_with_user4"
#
#     def test__if_change_works__single_user_and_single_group_change(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_single_user_and_single_group,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["users"]) == 1
#         assert rules[0]["users"][0]["username"] == "merge_requests_user1"
#         assert len(rules[0]["groups"]) == 1
#         assert rules[0]["groups"][0]["name"] == "gitlabform_tests_group_with_user4"
#
#         gf = GitLabForm(
#             config_string=config__approvers_single_user_and_single_group_change,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         assert len(rules[0]["users"]) == 1
#         assert rules[0]["users"][0]["username"] == "merge_requests_user2"
#         assert len(rules[0]["groups"]) == 1
#         assert (
#             rules[0]["groups"][0]["name"]
#             == "gitlabform_tests_group_with_user1_and_user2"
#         )
#
#     def test__more_than_one_user(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_more_than_one_user,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#
#         rule = rules[0]
#         assert len(rule["users"]) == 2
#         usernames_in_rule = set([user["username"] for user in rule["users"]])
#         assert usernames_in_rule == {"merge_requests_user1", "merge_requests_user2"}
#
#         assert len(rule["groups"]) == 0
#
#     def test__more_than_one_user_and_more_than_one_group(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_more_than_one_user_and_more_than_one_group,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#
#         rule = rules[0]
#         assert len(rule["users"]) == 2
#         usernames_in_rule = set([user["username"] for user in rule["users"]])
#         assert usernames_in_rule == {"merge_requests_user1", "merge_requests_user2"}
#
#         assert len(rule["groups"]) == 2
#         groupnames_in_rule = set([group["name"] for group in rule["groups"]])
#         assert groupnames_in_rule == {
#             "gitlabform_tests_group_with_user3",
#             "gitlabform_tests_group_with_user4",
#         }
#
#         assert len(rule["eligible_approvers"]) == 4
#         eligibleapprovers_in_rule = set(
#             [user["username"] for user in rule["eligible_approvers"]]
#         )
#         assert eligibleapprovers_in_rule == {
#             "merge_requests_user1",
#             "merge_requests_user2",
#             "merge_requests_user3",
#             "merge_requests_user4",
#         }
#
#     def test__more_than_one_user_and_more_than_one_group__change(self, gitlab):
#         gf = GitLabForm(
#             config_string=config__approvers_more_than_one_user_and_more_than_one_group,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#
#         rule = rules[0]
#         assert len(rule["users"]) == 2
#         usernames_in_rule = set([user["username"] for user in rule["users"]])
#         assert usernames_in_rule == {"merge_requests_user1", "merge_requests_user2"}
#
#         assert len(rule["groups"]) == 2
#         groupnames_in_rule = set([group["name"] for group in rule["groups"]])
#         assert groupnames_in_rule == {
#             "gitlabform_tests_group_with_user3",
#             "gitlabform_tests_group_with_user4",
#         }
#
#         assert len(rule["eligible_approvers"]) == 4
#         eligibleapprovers_in_rule = set(
#             [user["username"] for user in rule["eligible_approvers"]]
#         )
#         assert eligibleapprovers_in_rule == {
#             "merge_requests_user1",
#             "merge_requests_user2",
#             "merge_requests_user3",
#             "merge_requests_user4",
#         }
#
#         gf = GitLabForm(
#             config_string=config__approvers_more_than_one_user,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#
#         rule = rules[0]
#         assert len(rule["users"]) == 2
#         usernames_in_rule = set([user["username"] for user in rule["users"]])
#         assert usernames_in_rule == {"merge_requests_user1", "merge_requests_user2"}
#
#         assert len(rule["groups"]) == 0
#
#     def test__removing_preexisting_approvals(self, gitlab):
#         # add some preexisting approval rules
#         gitlab.create_approval_rule(
#             GROUP_AND_PROJECT_NAME,
#             "additional approval rule",
#             1,
#             ["merge_requests_user2"],
#             [GROUP_WITH_USER4],
#         )
#
#         gf = GitLabForm(
#             config_string=config__approvers_removing_preexisting_approvals,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         rules = gitlab.get_approvals_rules(GROUP_AND_PROJECT_NAME)
#         assert len(rules) == 1
#         rule = rules[0]
#         assert len(rule["users"]) == 1
#         usernames_in_rule = set([user["username"] for user in rule["users"]])
#         assert usernames_in_rule == {"merge_requests_user1"}
#
#         assert len(rule["groups"]) == 1
#         groupnames_in_rule = set([group["name"] for group in rule["groups"]])
#         assert groupnames_in_rule == {
#             "gitlabform_tests_group_with_user3",
#         }



import os
import textwrap

from logging import debug

from abc import ABC
from cli_ui import debug as verbose
from cli_ui import fatal

import yaml
from pathlib import Path
from mergedeep import merge
from gitlabform import EXIT_INVALID_INPUT


class ConfigurationCore(ABC):

    config = None
    config_dir = None

    def __init__(self, config_path=None, config_string=None):

        if config_path and config_string:
            fatal(
                "Please initialize with either config_path or config_string, not both.",
                exit_code=EXIT_INVALID_INPUT,
            )

        try:
            if config_string:
                verbose("Reading config from provided string.")
                self.config = yaml.safe_load(textwrap.dedent(config_string))
                self.config_dir = "."
            else:  # maybe config_path
                if "APP_HOME" in os.environ:
                    # using this env var should be considered unofficial, we need this temporarily
                    # for backwards compatibility. support for it may be removed without notice, do not use it!
                    config_path = os.path.join(os.environ["APP_HOME"], "config.yml")
                elif not config_path:
                    # this case is only meant for using gitlabform as a library
                    config_path = os.path.join(
                        str(Path.home()), ".gitlabform", "config.yml"
                    )
                elif config_path in [os.path.join(".", "config.yml"), "config.yml"]:
                    # provided points to config.yml in the app current working dir
                    config_path = os.path.join(os.getcwd(), "config.yml")

                verbose(f"Reading config from file: {config_path}")

                with open(config_path, "r") as ymlfile:
                    self.config = yaml.safe_load(ymlfile)
                    debug("Config parsed successfully as YAML.")

                # we need config path for accessing files for relative paths
                self.config_dir = os.path.dirname(config_path)

                if self.config.get("example_config"):
                    fatal(
                        "Example config detected, aborting.\n"
                        "Haven't you forgotten to use `-c <config_file>` parameter?\n"
                        "If you created your config based on the example config.yml,"
                        " then please remove 'example_config' key.",
                        exit_code=EXIT_INVALID_INPUT,
                    )

                if self.config.get("config_version", 1) != 2:
                    fatal(
                        "This version of GitLabForm requires 'config_version: 2' entry in the config.\n"
                        "This ensures that when the application behavior changes in a backward incompatible way,"
                        " you won't apply unexpected configuration to your GitLab instance.\n"
                        "Please read the upgrading guide here: https://bit.ly/3ub1g5C\n",
                        exit_code=EXIT_INVALID_INPUT,
                    )

                # we are NOT checking for the existence of non-empty 'projects_and_groups' key here
                # as it would break using GitLabForm as a library

        except (FileNotFoundError, IOError):
            raise ConfigFileNotFoundException(config_path)

        except Exception as e:
            raise ConfigInvalidException(e)

    def get(self, path, default=None):
        """
        :param path: "path" to given element in YAML file, for example for:

        group_settings:
          sddc:
            deploy_keys:
              qa_puppet:
                key: some key...
                title: some title...
                can_push: false

        ..a path to a single element array ['qa_puppet'] will be: "group_settings|sddc|deploy_keys".

        To get the dict under it use: get("group_settings|sddc|deploy_keys")

        :param default: the value to return if the key is not found. The default 'None' means that an exception
                        will be raised in such case.
        :return: element from YAML file (dict, array, string...)
        """
        tokens = path.split("|")
        current = self.config

        try:
            for token in tokens:
                current = current[token]
        except:
            if default is not None:
                return default
            else:
                raise KeyNotFoundException

        return current

    def get_group_config(self, group) -> dict:
        pass

    def get_project_config(self, group_and_project) -> dict:
        pass

    def get_common_config(self) -> dict:
        """
        :return: literal common configuration or empty dict if not defined
        """
        return self.get("projects_and_groups|*", {})

    def is_project_skipped(self, project) -> bool:
        pass

    def is_group_skipped(self, group) -> bool:
        pass

    def is_skipped(self, an_array: list, item: str) -> bool:
        pass

    @staticmethod
    def merge_configs(more_general_config, more_specific_config) -> dict:
        """
        :return: merge more general config with more specific configs.
                 More specific config values take precedence over more general ones.
        """
        return dict(merge({}, more_general_config, more_specific_config))


class ConfigFileNotFoundException(Exception):
    pass


class ConfigInvalidException(Exception):
    def __init__(self, underlying: Exception):
        self.underlying = underlying


class KeyNotFoundException(Exception):
    pass



from logging import debug
from cli_ui import debug as verbose
from cli_ui import warning

import copy
import textwrap
import yaml

from gitlabform.gitlab import GitLab
from gitlabform.gitlab.core import NotFoundException
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.difference_logger import hide


class SecretVariablesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("secret_variables", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        if (
            self.gitlab.get_project_settings(project_and_group)["builds_access_level"]
            == "disabled"
        ):
            warning(
                "Builds disabled in this project so I can't set secret variables here."
            )
            return

        debug(
            "Secret variables BEFORE: %s",
            self.gitlab.get_secret_variables(project_and_group),
        )

        for secret_variable in sorted(configuration["secret_variables"]):

            if "delete" in configuration["secret_variables"][secret_variable]:
                key = configuration["secret_variables"][secret_variable]["key"]
                if configuration["secret_variables"][secret_variable]["delete"]:
                    verbose(
                        f"Deleting {secret_variable}: {key} in project {project_and_group}"
                    )
                    try:
                        self.gitlab.delete_secret_variable(project_and_group, key)
                    except:
                        warning(
                            f"Could not delete variable {key} in group {project_and_group}"
                        )
                    continue

            verbose(f"Setting secret variable: {secret_variable}")
            try:
                self.gitlab.put_secret_variable(
                    project_and_group,
                    configuration["secret_variables"][secret_variable],
                )
            except NotFoundException:
                self.gitlab.post_secret_variable(
                    project_and_group,
                    configuration["secret_variables"][secret_variable],
                )

        debug(
            "Secret variables AFTER: %s",
            self.gitlab.get_secret_variables(project_and_group),
        )

    def _print_diff(self, project_and_group: str, configuration):

        try:
            current_secret_variables = self.gitlab.get_secret_variables(
                project_and_group
            )

            for secret_variable in current_secret_variables:
                secret_variable["value"] = hide(secret_variable["value"])

            verbose(f"Secret variables for {project_and_group} in GitLab:")

            verbose(
                textwrap.indent(
                    yaml.dump(current_secret_variables, default_flow_style=False),
                    "  ",
                )
            )
        except:
            verbose(
                f"Secret variables for {project_and_group} in GitLab cannot be checked."
            )

        verbose(f"Secret variables in {project_and_group} in configuration:")

        configured_secret_variables = copy.deepcopy(configuration)
        for key in configured_secret_variables.keys():
            configured_secret_variables[key]["value"] = hide(
                configured_secret_variables[key]["value"]
            )

        verbose(
            textwrap.indent(
                yaml.dump(configured_secret_variables, default_flow_style=False),
                "  ",
            )
        )



# import pytest
#
# from gitlabform.gitlab.core import UnexpectedResponseException
# from gitlabform.gitlabform import GitLabForm
# from gitlabform.gitlabform.test import (
#     create_group,
#     create_project_in_group,
#     get_gitlab,
#     GROUP_NAME,
# )
#
# PROJECT_NAME = "secret_variables_project"
# GROUP_AND_PROJECT_NAME = GROUP_NAME + "/" + PROJECT_NAME
#
#
# @pytest.fixture(scope="function")
# def gitlab(request):
#     create_group(GROUP_NAME)
#     create_project_in_group(GROUP_NAME, PROJECT_NAME)
#
#     gl = get_gitlab()
#
#     def fin():
#         gl.delete_project(GROUP_AND_PROJECT_NAME)
#
#     request.addfinalizer(fin)
#     return gl  # provide fixture value
#
#
# config_builds_not_enabled = f"""
# project_settings:
#   {GROUP_AND_PROJECT_NAME}:
#     project_settings:
#       builds_access_level: disabled
#     secret_variables:
#       foo:
#         key: FOO
#         value: 123
# """
#
# config_single_secret_variable = f"""
# project_settings:
#   {GROUP_AND_PROJECT_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     secret_variables:
#       foo:
#         key: FOO
#         value: 123
# """
#
# config_delete_secret_variable = f"""
# project_settings:
#   {GROUP_AND_PROJECT_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     secret_variables:
#       foo:
#         key: FOO
#         value: 123
#         delete: true
# """
#
# config_single_secret_variable2 = f"""
# project_settings:
#   {GROUP_AND_PROJECT_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     secret_variables:
#       foo:
#         key: FOO
#         value: 123456
# """
#
# config_more_secret_variables = f"""
# project_settings:
#   {GROUP_AND_PROJECT_NAME}:
#     project_settings:
#       builds_access_level: enabled
#     secret_variables:
#       foo:
#         key: FOO
#         value: 123456
#       bar:
#         key: BAR
#         value: bleble
# """
#
#
# class TestSecretVariables:
#     def test__builds_disabled(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_builds_not_enabled,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         with pytest.raises(UnexpectedResponseException):
#             # secret variables will NOT be available without builds_access_level in ['private', 'enabled']
#             gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#
#     def test__single_secret_variable(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#         assert len(secret_variables) == 1
#
#     def test__delete_secret_variable(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#         assert len(secret_variables) == 1
#         assert secret_variables[0]["value"] == "123"
#
#         gf = GitLabForm(
#             config_string=config_delete_secret_variable,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#         assert len(secret_variables) == 0
#
#     def test__reset_single_secret_variable(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_single_secret_variable,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#         assert len(secret_variables) == 1
#         assert secret_variables[0]["value"] == "123"
#
#         gf = GitLabForm(
#             config_string=config_single_secret_variable2,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#         assert len(secret_variables) == 1
#         assert secret_variables[0]["value"] == "123456"
#
#     def test__more_secret_variables(self, gitlab):
#         gf = GitLabForm(
#             config_string=config_more_secret_variables,
#             project_or_group=GROUP_AND_PROJECT_NAME,
#         )
#         gf.main()
#
#         secret_variables = gitlab.get_secret_variables(GROUP_AND_PROJECT_NAME)
#         secret_variables_keys = set([secret["key"] for secret in secret_variables])
#         assert len(secret_variables) == 2
#         assert secret_variables_keys == {"FOO", "BAR"}



import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
)


@pytest.fixture(scope="function")
def one_maintainer_and_two_developers(gitlab, group_and_project, users):

    gitlab.add_member_to_project(
        group_and_project, users[0], AccessLevel.MAINTAINER.value
    )
    gitlab.add_member_to_project(
        group_and_project, users[1], AccessLevel.DEVELOPER.value
    )
    gitlab.add_member_to_project(
        group_and_project, users[2], AccessLevel.DEVELOPER.value
    )

    yield group_and_project

    # we try to remove all users, not just the 3 added above,
    # on purpose, as more may have been added in the tests
    for user in users:
        gitlab.remove_member_from_project(group_and_project, user)


@pytest.fixture(scope="function")
def other_group_with_users(gitlab, other_group, other_users):
    gitlab.add_member_to_group(other_group, other_users[0], AccessLevel.OWNER.value)
    gitlab.add_member_to_group(other_group, other_users[1], AccessLevel.DEVELOPER.value)
    gitlab.remove_member_from_group(other_group, "root")

    # TODO: make it nicer than a duplicated list of users from above
    yield other_group, [other_users[0], other_users[1]]

    # we are running tests with root's token, so every group is created
    # with a single user - root as owner. we restore the group to
    # this state here.
    gitlab.add_member_to_group(other_group, "root", AccessLevel.OWNER.value)

    # we try to remove all users, not just those added above,
    # on purpose, as more may have been added in the tests
    for user in other_users:
        gitlab.remove_member_from_group(other_group, user)


class TestMembers:
    def test__add_user(
        self, gitlab, group_and_project, users, one_maintainer_and_two_developers
    ):
        members_before = gitlab.get_project_members(group_and_project)
        no_of_members_before = len(members_before)
        members_usernames_before = [member["username"] for member in members_before]

        user_to_add = users[3]
        assert user_to_add not in members_usernames_before

        add_users = f"""
        projects_and_groups:
          {group_and_project}:
            members:
              users:
                {user_to_add}: # new user
                  access_level: {AccessLevel.DEVELOPER.value}
        """

        run_gitlabform(add_users, group_and_project)

        members = gitlab.get_project_members(group_and_project)
        assert len(members) == no_of_members_before + 1

        members_usernames = [member["username"] for member in members]
        assert user_to_add in members_usernames

    # TODO: fix flaky test
    @pytest.mark.xfail(strict=False)
    def test__add_group(
        self,
        gitlab,
        group_and_project,
        one_maintainer_and_two_developers,
        other_group_with_users,
    ):
        other_group, other_group_users = other_group_with_users

        no_of_members_before = len(gitlab.get_project_members(group_and_project))
        no_of_members_of_group = len(other_group_users)

        # print(f"members before = {gitlab.get_project_members(group_and_project)}")
        # print(f"members of the group = {other_group_users}")

        no_of_groups_shared_before = len(
            gitlab.get_shared_with_groups(group_and_project)
        )
        assert no_of_groups_shared_before == 0

        add_group = f"""
        projects_and_groups:
          {group_and_project}:
            members:
              groups:
                {other_group}:
                  group_access: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(add_group, group_and_project)

        members = gitlab.get_project_members(group_and_project, all=True)

        # TODO: the +1 is for root, but actually why is that user inherited here?
        assert len(members) == no_of_members_before + no_of_members_of_group + 1

        for member in members:
            if member["username"] in other_group_users:
                # "group_access" is the *maximum* access level, see
                # https://docs.gitlab.com/ee/user/project/members/share_project_with_groups.html#maximum-access-level
                assert member["access_level"] <= AccessLevel.MAINTAINER.value

        no_of_groups_shared = len(gitlab.get_shared_with_groups(group_and_project))
        assert no_of_groups_shared == 1

    def test__no_groups_and_no_users(self, gitlab, group, project):

        config_with_error = f"""
        projects_and_groups:
          {group}/{project}:
            members:
              # there should be a sub-key 'users' here, not directly a user
              {project}_user1: 
                access_level: {AccessLevel.DEVELOPER.value}
        """

        with pytest.raises(SystemExit):
            run_gitlabform(config_with_error, f"{group}/{project}")

    def test__enforce(
        self, gitlab, group_and_project, users, one_maintainer_and_two_developers
    ):
        members_before = gitlab.get_project_members(group_and_project)
        members_usernames_before = [member["username"] for member in members_before]

        user_to_add = users[3]
        assert user_to_add not in members_usernames_before

        enforce_users = f"""
            projects_and_groups:
              {group_and_project}:
                members:
                  users:
                    {user_to_add}: # new user
                      access_level: {AccessLevel.MAINTAINER.value}
                  enforce: true
            """

        run_gitlabform(enforce_users, group_and_project)

        members = gitlab.get_project_members(group_and_project)

        assert len(members) == 1
        assert members[0]["username"] == user_to_add



import pytest

from gitlabform.gitlab import AccessLevel
from tests.acceptance import (
    run_gitlabform,
    DEFAULT_README,
    get_gitlab,
)


gl = get_gitlab()


@pytest.fixture(scope="function")
def branches(request, gitlab, group_and_project):
    branches = [
        "protect_branch_but_allow_all",
        "protect_branch_with_code_owner_approval_required",
        "protect_branch_and_disallow_all",
        "protect_branch_and_allow_merges",
        "protect_branch_and_allow_pushes",
        "protect_branch_and_allow_merges_access_levels",
        "protect_branch_and_allow_pushes_access_levels",
        "protect_branch_and_allowed_to_push",
        "protect_branch_and_allowed_to_merge",
        "protect_branch_and_allow_access_levels_with_user_ids",
        "protect_branch",
    ]
    for branch in branches:
        gitlab.create_branch(group_and_project, branch, "main")

    def fin():
        for branch in branches:
            gitlab.delete_branch(group_and_project, branch)

        gitlab.set_file(
            group_and_project,
            "main",
            "README.md",
            DEFAULT_README,
            "Reset default content",
        )

    request.addfinalizer(fin)


@pytest.fixture(scope="function")
def one_maintainer_and_two_developers(gitlab, group_and_project, users):

    gitlab.add_member_to_project(
        group_and_project, users[0], AccessLevel.MAINTAINER.value
    )
    gitlab.add_member_to_project(
        group_and_project, users[1], AccessLevel.DEVELOPER.value
    )
    gitlab.add_member_to_project(
        group_and_project, users[2], AccessLevel.DEVELOPER.value
    )

    yield group_and_project

    # we try to remove all users, not just the 3 added above,
    # on purpose, as more may have been added in the tests
    for user in users:
        gitlab.remove_member_from_project(group_and_project, user)


class TestBranches:
    def test__protect_branch_but_allow_all(self, gitlab, group_and_project, branches):

        protect_branch_but_allow_all = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_but_allow_all:
                protected: true
                developers_can_push: true
                developers_can_merge: true
        """
        run_gitlabform(protect_branch_but_allow_all, group_and_project)
        branch = gitlab.get_branch(group_and_project, "protect_branch_but_allow_all")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is True
        assert branch["developers_can_merge"] is True

    # @pytest.mark.skipif(
    #     gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    # )
    # def test__code_owners_approval(self, gitlab, group_and_project, branches):
    #     group_and_project = group_and_project
    #
    #     branch_access_levels = gitlab.get_branch_access_levels(
    #         group_and_project, "protect_branch_but_allow_all"
    #     )
    #     assert branch_access_levels["code_owner_approval_required"] is False
    #
    #     protect_branch_with_code_owner_approval_required = f"""
    #     projects_and_groups:
    #       {group_and_project}:
    #         branches:
    #           protect_branch_with_code_owner_approval_required:
    #             protected: true
    #             developers_can_push: false
    #             developers_can_merge: true
    #             code_owner_approval_required: true
    #     """
    #
    #     run_gitlabform(
    #         protect_branch_with_code_owner_approval_required, group_and_project
    #     )
    #
    #     branch_access_levels = gitlab.get_branch_access_levels(
    #         group_and_project, "protect_branch_with_code_owner_approval_required"
    #     )
    #     assert branch_access_levels["code_owner_approval_required"] is True

    def test__protect_branch_and_disallow_all(
        self, gitlab, group_and_project, branches
    ):

        protect_branch_and_disallow_all = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_disallow_all:
                protected: true
                developers_can_push: false
                developers_can_merge: false
        """
        run_gitlabform(protect_branch_and_disallow_all, group_and_project)
        branch = gitlab.get_branch(group_and_project, "protect_branch_and_disallow_all")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is False
        assert branch["developers_can_merge"] is False

    def test__mixed_config(self, gitlab, group_and_project, branches):

        mixed_config = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allow_merges:
                protected: true
                developers_can_push: false
                developers_can_merge: true
              protect_branch_and_allow_pushes:
                protected: true
                developers_can_push: true
                developers_can_merge: false
        """
        run_gitlabform(mixed_config, group_and_project)
        branch = gitlab.get_branch(group_and_project, "protect_branch_and_allow_merges")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is False
        assert branch["developers_can_merge"] is True

        branch = gitlab.get_branch(group_and_project, "protect_branch_and_allow_pushes")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is True
        assert branch["developers_can_merge"] is False

        unprotect_branches = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allow_merges:
                protected: false
              protect_branch_and_allow_pushes:
                protected: false
        """
        run_gitlabform(unprotect_branches, group_and_project)

        for branch in [
            "protect_branch_and_allow_merges",
            "protect_branch_and_allow_pushes",
        ]:
            branch = gitlab.get_branch(group_and_project, branch)
            assert branch["protected"] is False

    def test__mixed_config_with_new_api(
        self,
        gitlab,
        group_and_project,
        branches,
        users,
        one_maintainer_and_two_developers,
    ):
        mixed_config_with_access_levels = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allow_merges_access_levels:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.DEVELOPER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
              '*_allow_pushes_access_levels':
                protected: true
                push_access_level: {AccessLevel.DEVELOPER.value}
                merge_access_level: {AccessLevel.DEVELOPER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(mixed_config_with_access_levels, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "protect_branch_and_allow_merges_access_levels"
        )
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.DEVELOPER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "*_allow_pushes_access_levels"
        )
        assert push_access_levels == [AccessLevel.DEVELOPER.value]
        assert merge_access_levels == [AccessLevel.DEVELOPER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        mixed_config_with_access_levels_update = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allow_merges_access_levels:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
              '*_allow_pushes_access_levels':
                protected: true
                push_access_level: {AccessLevel.MAINTAINER.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(mixed_config_with_access_levels_update, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "protect_branch_and_allow_merges_access_levels"
        )
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "*_allow_pushes_access_levels"
        )
        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        mixed_config_with_access_levels_unprotect_branches = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allow_merges_access_levels:
                protected: false
              '*_allow_pushes_access_levels':
                protected: false
        """

        run_gitlabform(
            mixed_config_with_access_levels_unprotect_branches, group_and_project
        )

        for branch in [
            "protect_branch_and_allow_merges_access_levels",
            "protect_branch_and_allow_pushes_access_levels",
        ]:
            branch = gitlab.get_branch(group_and_project, branch)
            assert branch["protected"] is False

    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test__allow_user_ids(
        self,
        gitlab,
        group_and_project,
        branches,
        users,
        one_maintainer_and_two_developers,
    ):

        user_allowed_to_push_id = gitlab.get_user_to_protect_branch(users[0])
        user_allowed_to_merge_id = gitlab.get_user_to_protect_branch(users[1])
        user_allowed_to_push_and_allowed_to_merge_id = (
            gitlab.get_user_to_protect_branch(users[2])
        )

        # testing allowed_to_push  and allowed_to_merge for user support on protect branch (gitlab premium feature)
        mixed_config_with_allowed_to_push_and_merge = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allowed_to_merge:
                protected: true
                allowed_to_push:
                  - access_level: {AccessLevel.NO_ACCESS.value} 
                allowed_to_merge:
                  - access_level: {AccessLevel.DEVELOPER.value} 
                  - user_id: {user_allowed_to_merge_id}
                  - user: {users[2]}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
              '*_and_allowed_to_push':
                protected: true
                allowed_to_push:
                  - access_level: {AccessLevel.DEVELOPER.value} 
                  - user_id: {user_allowed_to_push_id}
                  - user: {users[1]} 
                allowed_to_merge:
                  - access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.DEVELOPER.value}
        """

        run_gitlabform(mixed_config_with_allowed_to_push_and_merge, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "protect_branch_and_allowed_to_merge"
        )

        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.DEVELOPER.value]

        current_push_access_user_ids = []
        current_push_access_user_ids.sort()
        assert push_access_user_ids == current_push_access_user_ids

        current_merge_access_user_ids = [
            user_allowed_to_merge_id,
            user_allowed_to_push_and_allowed_to_merge_id,
        ]
        current_merge_access_user_ids.sort()
        assert merge_access_user_ids == current_merge_access_user_ids

        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "*_and_allowed_to_push"
        )

        assert push_access_levels == [AccessLevel.DEVELOPER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]

        current_push_access_user_ids = [
            user_allowed_to_push_id,
            user_allowed_to_merge_id,
        ]
        current_push_access_user_ids.sort()
        assert push_access_user_ids == current_push_access_user_ids

        current_merge_access_user_ids = []
        current_merge_access_user_ids.sort()
        assert merge_access_user_ids == current_merge_access_user_ids

        assert unprotect_access_level is AccessLevel.DEVELOPER.value

        mixed_config_with_allowed_to_push_and_merge_update = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allowed_to_merge:
                protected: true
                allowed_to_push:
                  - access_level: {AccessLevel.NO_ACCESS.value} 
                allowed_to_merge:
                  - access_level: {AccessLevel.MAINTAINER.value} 
                  - user_id: {user_allowed_to_merge_id}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
              '*_and_allowed_to_push':
                protected: true
                allowed_to_push:
                  - access_level: {AccessLevel.MAINTAINER.value} 
                  - user_id: {user_allowed_to_push_id}
                  - user: {users[2]}
                  - user: {users[1]} 
                allowed_to_merge:
                  - access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(
            mixed_config_with_allowed_to_push_and_merge_update, group_and_project
        )

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "protect_branch_and_allowed_to_merge"
        )

        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]

        current_push_access_user_ids = []
        current_push_access_user_ids.sort()

        assert push_access_user_ids == current_push_access_user_ids

        current_merge_access_user_ids = [user_allowed_to_merge_id]
        current_merge_access_user_ids.sort()
        assert merge_access_user_ids == current_merge_access_user_ids
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project, "*_and_allowed_to_push"
        )

        assert push_access_levels == [AccessLevel.MAINTAINER.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]

        current_push_access_user_ids = [
            user_allowed_to_push_id,
            user_allowed_to_merge_id,
            user_allowed_to_push_and_allowed_to_merge_id,
        ]
        current_push_access_user_ids.sort()

        assert push_access_user_ids == current_push_access_user_ids

        current_merge_access_user_ids = []
        current_merge_access_user_ids.sort()
        assert merge_access_user_ids == current_merge_access_user_ids
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        mixed_config_with_allow_access_levels_with_user_ids = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch_and_allow_access_levels_with_user_ids:
                protected: true
                push_access_level: {AccessLevel.DEVELOPER.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                allowed_to_push:
                  - access_level: {AccessLevel.MAINTAINER.value} 
                  - user_id: {user_allowed_to_push_id}
                  - user: {users[2]}
                allowed_to_merge:
                  - access_level: {AccessLevel.DEVELOPER.value} 
                  - user_id: {user_allowed_to_merge_id}
                  - user: {users[0]}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(
            mixed_config_with_allow_access_levels_with_user_ids, group_and_project
        )

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(
            group_and_project,
            "protect_branch_and_allow_access_levels_with_user_ids",
        )

        assert push_access_levels == [
            AccessLevel.DEVELOPER.value,
            AccessLevel.MAINTAINER.value,
        ]
        assert merge_access_levels == [
            AccessLevel.DEVELOPER.value,
            AccessLevel.MAINTAINER.value,
        ]

        current_push_access_user_ids = [
            user_allowed_to_push_id,
            user_allowed_to_push_and_allowed_to_merge_id,
        ]
        current_push_access_user_ids.sort()

        assert push_access_user_ids == current_push_access_user_ids

        current_merge_access_user_ids = [
            user_allowed_to_merge_id,
            user_allowed_to_push_id,
        ]
        current_merge_access_user_ids.sort()
        assert merge_access_user_ids == current_merge_access_user_ids
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

    def test_protect_branch_with_old_api_next_update_with_new_api_and_unprotect(
        self, gitlab, group_and_project, branches
    ):

        config_protect_branch_with_old_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                developers_can_push: true
                developers_can_merge: true
        """

        run_gitlabform(config_protect_branch_with_old_api, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is True
        assert branch["developers_can_merge"] is True

        config_protect_branch_with_new_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_protect_branch_with_new_api, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "protect_branch")
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        config_protect_branch_unprotect = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: false
        """

        run_gitlabform(config_protect_branch_unprotect, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is False

    def test_protect_branch_with_new_api_next_update_with_old_api_and_unprotect(
        self, gitlab, group_and_project, branches
    ):

        config_protect_branch_with_new_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_protect_branch_with_new_api, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "protect_branch")
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        config_protect_branch_with_old_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                developers_can_push: true
                developers_can_merge: true
        """

        run_gitlabform(config_protect_branch_with_old_api, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is True
        assert branch["developers_can_merge"] is True

        config_protect_branch_unprotect = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: false
        """

        run_gitlabform(config_protect_branch_unprotect, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is False

    @pytest.mark.skipif(
        gl.has_no_license(), reason="this test requires a GitLab license (Paid/Trial)"
    )
    def test_protect_branch_with_old_api_next_update_with_new_api_and_userid_and_unprotect(
        self,
        gitlab,
        group_and_project,
        branches,
        users,
        one_maintainer_and_two_developers,
    ):

        config_protect_branch_with_old_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                developers_can_push: true
                developers_can_merge: true
        """

        run_gitlabform(config_protect_branch_with_old_api, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is True
        assert branch["developers_can_merge"] is True

        user_allowed_to_push_id = gitlab.get_user_to_protect_branch(users[0])
        user_allowed_to_merge_id = gitlab.get_user_to_protect_branch(users[1])

        config_protect_branch_with_new_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                push_access_level: {AccessLevel.DEVELOPER.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                allowed_to_push:
                   - access_level: {AccessLevel.MAINTAINER.value} 
                   - user_id: {user_allowed_to_push_id}
                   - user: {users[1]}
                allowed_to_merge:
                   - access_level: {AccessLevel.MAINTAINER.value} 
                   - user_id: {user_allowed_to_merge_id}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_protect_branch_with_new_api, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "protect_branch")

        assert push_access_levels == [
            AccessLevel.DEVELOPER.value,
            AccessLevel.MAINTAINER.value,
        ]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]

        current_push_access_user_ids = [
            user_allowed_to_push_id,
            user_allowed_to_merge_id,
        ]
        current_push_access_user_ids.sort()
        assert push_access_user_ids == current_push_access_user_ids

        current_merge_access_user_ids = [user_allowed_to_merge_id]
        current_merge_access_user_ids.sort()
        assert merge_access_user_ids == current_merge_access_user_ids

        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        config_protect_branch_unprotect = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: false
        """

        run_gitlabform(config_protect_branch_unprotect, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is False

    def test_unprotect_when_the_rest_of_the_parameters_are_still_specified_old_api(
        self, gitlab, group_and_project, branches
    ):

        config_protect_branch_with_old_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                developers_can_push: true
                developers_can_merge: true
        """

        run_gitlabform(config_protect_branch_with_old_api, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is True
        assert branch["developers_can_push"] is True
        assert branch["developers_can_merge"] is True

        config_unprotect_branch_with_old_api = f"""
        gitlab:
          api_version: 4

        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: false
                developers_can_push: true
                developers_can_merge: true
        """

        run_gitlabform(config_unprotect_branch_with_old_api, group_and_project)

        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is False

    def test_unprotect_when_the_rest_of_the_parameters_are_still_specified_new_api(
        self, gitlab, group_and_project, branches
    ):

        config_protect_branch_with_new_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: true
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_protect_branch_with_new_api, group_and_project)

        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "protect_branch")
        assert push_access_levels == [AccessLevel.NO_ACCESS.value]
        assert merge_access_levels == [AccessLevel.MAINTAINER.value]
        assert push_access_user_ids == []
        assert merge_access_user_ids == []
        assert unprotect_access_level is AccessLevel.MAINTAINER.value

        config_unprotect_branch_with_new_api = f"""
        projects_and_groups:
          {group_and_project}:
            branches:
              protect_branch:
                protected: false
                push_access_level: {AccessLevel.NO_ACCESS.value}
                merge_access_level: {AccessLevel.MAINTAINER.value}
                unprotect_access_level: {AccessLevel.MAINTAINER.value}
        """

        run_gitlabform(config_unprotect_branch_with_new_api, group_and_project)

        # old API
        branch = gitlab.get_branch(group_and_project, "protect_branch")
        assert branch["protected"] is False

        # new API
        (
            push_access_levels,
            merge_access_levels,
            push_access_user_ids,
            merge_access_user_ids,
            unprotect_access_level,
        ) = gitlab.get_only_branch_access_levels(group_and_project, "protect_branch")
        assert push_access_levels is None
        assert merge_access_levels is None
        assert push_access_user_ids is None
        assert merge_access_user_ids is None
        assert unprotect_access_level is None



import codecs
import os
import re
from setuptools import setup, find_packages

with codecs.open("README.md", encoding="utf-8") as f:
    readme = f.read()

# ToC doesn't work when viewed in PyPI :(
# it also contains link to changelog which for PyPI is added to the end of the readme
# so let's remove the ToC here
regexp = re.compile(r"## Table of Contents(.*)## Features", re.MULTILINE + re.DOTALL)
readme = re.sub(regexp, "## Features", readme)

with codecs.open("CHANGELOG.md", encoding="utf-8") as f:
    changelog = f.read()

long_description = readme + "\n" + changelog


def get_version_file_path():
    github_actions_path = "/home/runner/work/gitlabform/gitlabform"
    if os.path.isfile(github_actions_path + "/version"):
        return github_actions_path + "/version"
    else:
        return "version"


setup(
    name="gitlabform",
    version=open(get_version_file_path()).read(),
    description='Specialized "configuration as a code" tool for GitLab projects, groups and more'
    " using hierarchical configuration written in YAML",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/egnyte/gitlabform",
    author="Egnyte and GitHub Contributors",
    keywords=["gitlab", "configuration-as-code"],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Software Development :: Version Control :: Git",
    ],
    packages=find_packages(),
    install_requires=[
        "certifi",  # we want the latest root certs for security
        "requests==2.26.0",
        "Jinja2==2.11.3",
        "MarkupSafe==1.1.1",
        "PyYAML==6.0",
        "luddite==1.0.2",
        "cli-ui==0.15.2",
        "packaging==21.2",
        "mergedeep==1.3.4",
    ],
    extras_require={
        "test": [
            "pytest==6.2.5",
            "xkcdpass==1.19.3",
            "pre-commit==2.15.0",  # not really for tests, but for development
            "coverage==6.1.1",
            "pytest-cov==3.0.0",
        ],
    },
    entry_points={
        "console_scripts": [
            "gitlabform=gitlabform.run:run",
        ],
    },
)



import pytest

from tests.acceptance import (
    get_gitlab,
    create_group,
    create_groups,
    delete_groups,
    create_project,
    get_random_name,
    create_users,
    delete_users,
)


@pytest.fixture(scope="session")
def gitlab():
    gl = get_gitlab()
    yield gl  # provide fixture value


@pytest.fixture(scope="class")
def group_and_project(group, project):
    return f"{group}/{project}"


@pytest.fixture(scope="class")
def group():
    group_name = get_random_name()
    create_group(group_name)

    yield group_name

    gl = get_gitlab()
    gl.delete_group(group_name)


@pytest.fixture(scope="class")
def other_group():
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    group_name = get_random_name()
    create_group(group_name)

    yield group_name

    gl = get_gitlab()
    gl.delete_group(group_name)


@pytest.fixture(scope="class")
def sub_group(group):
    gl = get_gitlab()
    parent_id = gl.get_group_id_case_insensitive(group)
    group_name = get_random_name()
    create_group(group_name, parent_id)

    yield group + "/" + group_name

    gl = get_gitlab()
    gl.delete_group(group + "/" + group_name)


@pytest.fixture(scope="class")
def project(group):
    project_name = get_random_name()
    create_project(group, project_name)

    yield project_name

    gl = get_gitlab()
    gl.delete_project(f"{group}/{project_name}")


@pytest.fixture(scope="class")
def other_project(group):
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    project_name = get_random_name()
    create_project(group, project_name)

    yield project_name

    gl = get_gitlab()
    gl.delete_project(f"{group}/{project_name}")


@pytest.fixture(scope="class")
def groups(users):
    no_of_groups = 4

    group_name_base = get_random_name()
    groups = create_groups(group_name_base, no_of_groups)

    yield groups

    delete_groups(group_name_base, no_of_groups)


@pytest.fixture(scope="class")
def users(group):
    no_of_users = 4

    username_base = get_random_name()
    users = create_users(username_base, no_of_users)

    yield users

    delete_users(username_base, no_of_users)


@pytest.fixture(scope="class")
def other_users():
    # TODO: deduplicate this - it's a copy and paste from the above fixture
    no_of_users = 4

    username_base = get_random_name()
    users = create_users(username_base, no_of_users)

    yield users

    delete_users(username_base, no_of_users)



from logging import debug

from gitlabform.configuration.groups import ConfigurationGroups


class ConfigurationProjectsAndGroups(ConfigurationGroups):
    def __init__(self, config_path=None, config_string=None):
        super().__init__(config_path, config_string)

    def get_projects(self) -> list:
        """
        :return: sorted list of projects names, that are EXPLICITLY defined in the config
        """
        projects = []
        projects_and_groups = self.get("projects_and_groups")
        for element in projects_and_groups.keys():
            if element != "*" and not element.endswith("/*"):
                projects.append(element)
        return sorted(projects)

    def get_effective_config_for_project(self, group_and_project) -> dict:
        """
        :param group_and_project: "project_group/project_name"
        :return: merged configuration for this project, from common, group/subgroup and project level.
                 If project belongs to a subgroup, like "x/y/z", then it gets config from both group "x" as well
                 as subgroup "y".
                 Merging is additive.
        """

        common_config = self.get_common_config()
        debug("Common config: %s" % common_config)

        group, project = group_and_project.rsplit("/", 1)
        if "/" in group:
            group_config = self.get_effective_subgroup_config(group)
        else:
            group_config = self.get_group_config(group)
        debug("Effective group/subgroup config: %s" % group_config)

        project_config = self.get_project_config(group_and_project)
        debug("Project config: %s" % project_config)

        common_and_group_config = self.merge_configs(common_config, group_config)
        debug("Effective config common+group/subgroup: %s" % common_and_group_config)

        effective_config = self.merge_configs(common_and_group_config, project_config)
        debug("Effective config common+group/subgroup+project: %s" % effective_config)

        return effective_config


class EmptyConfigException(Exception):
    pass



from cli_ui import fatal

from gitlabform import EXIT_INVALID_INPUT, Groups, Projects
from gitlabform.gitlab.core import NotFoundException


class GroupsAndProjectsProvider:
    """
    For a query like "project/group", "group", "group/subgroup", ALL or ALL_DEFINED this
    class gets the effective lists of groups and project, taking into account skipped groups
    and projects and the fact that the group and project names case are somewhat case-sensitive.
    """

    def __init__(self, gitlab, configuration, include_archived_projects):
        self.gitlab = gitlab
        self.configuration = configuration
        self.include_archived_projects = include_archived_projects

    def get_groups_and_projects(self, target: str) -> (Groups, Projects):
        """
        :param target: "project/group", "group", "group/subgroup", ALL or ALL_DEFINED
        :return: tuple with Groups and Projects
        """

        if target not in ["ALL", "ALL_DEFINED"]:
            groups, projects = self._get_single_group_or_project(target)
            if len(groups.get_effective()) == 1:
                projects = self._get_projects(target, groups)
        else:
            groups = self._get_groups(target)
            projects = self._get_projects(target, groups)

        return groups, projects

    def _get_single_group_or_project(self, target: str) -> (Groups, Projects):

        groups = Groups()
        projects = Projects()

        # it may be a subgroup or a single group...
        try:
            maybe_group = self.gitlab.get_group_case_insensitive(target)
            groups.add_requested([maybe_group["full_path"]])

        except NotFoundException:

            # ...or a single project
            try:
                maybe_project = self.gitlab.get_project_case_insensitive(target)
                projects.add_requested([maybe_project["path_with_namespace"]])

            except NotFoundException:
                pass

        return groups, projects

    def _get_groups(self, target: str) -> Groups:

        groups = Groups()

        if target == "ALL":
            groups.add_requested(self.gitlab.get_groups())
        else:  # ALL_DEFINED
            groups.add_requested(self.configuration.get_groups())

        groups.add_omitted("skipped", self._get_skipped_groups(groups.get_effective()))

        if target == "ALL_DEFINED":
            # check if all the groups from the config actually exist
            self._verify_if_groups_exist(groups.get_effective())

        return groups

    def _get_projects(self, target: str, groups: Groups) -> Projects:

        projects = Projects()

        # the source of projects are the *effective* requested groups
        (
            projects_from_groups,
            archived_projects_from_groups,
        ) = self._get_all_and_archived_projects_from_groups(groups.get_effective())
        projects.add_requested(projects_from_groups)
        projects.add_omitted("archived", archived_projects_from_groups)

        if target == "ALL_DEFINED":

            # in this case we also need to get the list of projects explicitly
            # defined in the configuration

            projects_from_configuration = self.configuration.get_projects()

            # ...but we don't need to re-check for being archived projects that we
            # already got from groups

            # TODO: this check should be case-insensitive
            projects_from_configuration_not_from_groups = [
                project
                for project in projects_from_configuration
                if project not in projects.requested
            ]

            archived_projects_from_configuration_not_from_groups = (
                self._verify_if_projects_exist_and_get_archived_projects(
                    projects_from_configuration_not_from_groups
                )
            )

            projects.add_requested(projects_from_configuration_not_from_groups)
            projects.add_omitted(
                "archived", archived_projects_from_configuration_not_from_groups
            )

        # TODO: consider checking for skipped earlier to avoid making requests for projects that will be skipped anyway
        projects.add_omitted(
            "skipped", self._get_skipped_projects(projects.get_effective())
        )

        return projects

    def _verify_if_groups_exist(self, groups: list):
        for group in groups:
            try:
                self.gitlab.get_group_case_insensitive(group)
            except NotFoundException:
                fatal(
                    f"Configuration contains group {group} but it cannot be found in GitLab!",
                    exit_code=EXIT_INVALID_INPUT,
                )

    def _verify_if_projects_exist_and_get_archived_projects(
        self, projects: list
    ) -> list:
        archived = []
        for project in projects:
            try:
                project_object = self.gitlab.get_project_case_insensitive(project)
                if project_object["archived"]:
                    archived.append(project_object["path_with_namespace"])
            except NotFoundException:
                fatal(
                    f"Configuration contains project {project} but it cannot be found in GitLab!",
                    exit_code=EXIT_INVALID_INPUT,
                )
        return archived

    def _get_all_and_archived_projects_from_groups(self, groups: list) -> (list, list):
        all = []
        archived = []
        for group in groups:
            if self.include_archived_projects:
                all += self.gitlab.get_projects(group, include_archived=True)
            else:
                project_objects = self.gitlab.get_projects(
                    group, include_archived=True, only_names=False
                )
                for project_object in project_objects:
                    project = project_object["path_with_namespace"]
                    all.append(project)
                    if project_object["archived"]:
                        archived.append(project)

        # deduplicate as we may have a group X and its subgroup X/Y in the groups list so the effective projects
        # may occur more than once
        return list(set(all)), list(set(archived))

    def _get_skipped_groups(self, groups: list) -> list:
        skipped = []
        for group in groups:
            if self.configuration.is_group_skipped(group):
                skipped.append(group)

        return skipped

    def _get_skipped_projects(self, projects: list) -> list:
        skipped = []
        for project in projects:
            if self.configuration.is_project_skipped(project):
                skipped.append(project)

        return skipped



from logging import debug
from cli_ui import debug as verbose

from gitlabform.gitlab import GitLab
from gitlabform.processors.abstract_processor import AbstractProcessor
from gitlabform.processors.util.difference_logger import DifferenceLogger


class ProjectPushRulesProcessor(AbstractProcessor):
    def __init__(self, gitlab: GitLab):
        super().__init__("project_push_rules", gitlab)

    def _process_configuration(self, project_and_group: str, configuration: dict):
        push_rules = configuration["project_push_rules"]
        old_project_push_rules = self.gitlab.get_project_push_rules(project_and_group)
        debug("Project push rules settings BEFORE: %s", old_project_push_rules)
        if old_project_push_rules:
            verbose(f"Updating project push rules: {push_rules}")
            self.gitlab.put_project_push_rules(project_and_group, push_rules)
        else:
            verbose(f"Creating project push rules: {push_rules}")
            self.gitlab.post_project_push_rules(project_and_group, push_rules)
        debug(
            "Project push rules AFTER: %s",
            self.gitlab.get_project_push_rules(project_and_group),
        )

    def _print_diff(self, project_and_group: str, push_rules):
        current_push_rules = self.gitlab.get_project_push_rules(project_and_group)
        DifferenceLogger.log_diff(
            "Project %s push rules changes" % project_and_group,
            current_push_rules,
            push_rules,
        )
